@@ -29,11 +29,11 @@ describe("follow-redirects", function () {
     return {
       app: app,
       protocol: "https",
-      cert: fs.readFileSync(path.resolve(__dirname, "assets/TestServer.crt")),
-      key: fs.readFileSync(path.resolve(__dirname, "assets/TestServer.pem")),
+      cert: fs.readFileSync(path.resolve(__dirname, "assets/localhost.crt")),
+      key: fs.readFileSync(path.resolve(__dirname, "assets/localhost.key")),
     };
   }
-  var ca = fs.readFileSync(path.resolve(__dirname, "assets/TestCA.crt"));
+  var ca = fs.readFileSync(path.resolve(__dirname, "assets/ca.crt"));
 
   var app;
   var app2;
@@ -281,9 +281,58 @@ describe("follow-redirects", function () {
       .then(function (error) {
         assert(error instanceof Error);
         assert.equal(error.code, "ERR_FR_REDIRECTION_FAILURE");
-        assert.equal(error.message, "Redirected request failed: Request path contains unescaped characters");
         assert(error.cause instanceof Error);
-        assert.equal(error.cause.code, "ERR_UNESCAPED_CHARACTERS");
+        switch (error.cause.code) {
+        // Node 17+
+        case "ERR_INVALID_URL":
+          assert.equal(error.message, "Redirected request failed: Invalid URL");
+          break;
+        // Older Node versions
+        case "ERR_UNESCAPED_CHARACTERS":
+          assert.equal(error.message, "Redirected request failed: Request path contains unescaped characters");
+          break;
+        default:
+          throw new Error("Unexpected error code " + error.code);
+        }
+      });
+  });
+
+  it("emits an error whem url.resolve fails", function () {
+    app.get("/a", redirectsTo("/b"));
+    var urlResolve = url.resolve;
+    url.resolve = function () {
+      throw new Error();
+    };
+
+    return server.start(app)
+      .then(asPromise(function (resolve) {
+        http.get("http://localhost:3600/a").on("error", resolve);
+      }))
+      .then(function (error) {
+        url.resolve = urlResolve;
+        assert.equal(error.code, "ERR_FR_REDIRECTION_FAILURE");
+      });
+  });
+
+  it("emits an error when the request fails for another reason", function () {
+    app.get("/a", function (req, res) {
+      res.socket.write("HTTP/1.1 301 Moved Permanently\n");
+      res.socket.write("Location: other\n");
+      res.socket.write("\n");
+      res.socket.end();
+    });
+
+    return server.start(app)
+      .then(asPromise(function (resolve) {
+        var request = http.get("http://localhost:3600/a");
+        request._performRequest = function () {
+          throw new Error("custom");
+        };
+        request.on("error", resolve);
+      }))
+      .then(function (error) {
+        assert(error instanceof Error);
+        assert.equal(error.message, "Redirected request failed: custom");
       });
   });
 
@@ -334,6 +383,25 @@ describe("follow-redirects", function () {
         });
     });
 
+    it("handles errors occuring before a socket is established", function () {
+      app.get("/redirect", redirectsTo("http://localhost:3602/b"));
+
+      var req;
+      return server.start(app)
+        .then(asPromise(function (resolve, reject) {
+          req = http.get("http://localhost:3600/redirect", reject);
+          req.setTimeout(3000, function () {
+            throw new Error("should not have timed out");
+          });
+          req.emit("error", new Error());
+          req.on("error", resolve);
+        }))
+        .then(function (error) {
+          assert.equal(error.code, "ECONNREFUSED");
+          clock.tick(5000);
+        });
+    });
+
     it("sets a timeout when the socket already exists", function () {
       app.get("/timeout", delay(clock, 5000, sendsJson({ timed: "out" })));
 
@@ -353,6 +421,24 @@ describe("follow-redirects", function () {
         }));
     });
 
+    it("destroys the socket after configured inactivity period", function () {
+      app.get("/data", delay(clock, 3000, sendsJson({ took: "toolongtosenddata" })));
+
+      return server.start(app)
+        .then(asPromise(function (resolve, reject) {
+          var req = http.get("http://localhost:3600/data", concatJson(reject, reject));
+          req.on("error", reject);
+          req.setTimeout(100, function () {
+            throw new Error("should not have timed out");
+          });
+          req.on("socket", function () {
+            req.socket.on("timeout", function () {
+              resolve();
+            });
+          });
+        }));
+    });
+
     it("should timeout on the final request", function () {
       app.get("/redirect1", redirectsTo("/redirect2"));
       app.get("/redirect2", redirectsTo("/timeout"));
@@ -1160,7 +1246,7 @@ describe("follow-redirects", function () {
   });
 
   describe("when redirecting to a different host while the host header is set", function () {
-    it("uses the new host header", function () {
+    it("uses the new host header if redirect host is different", function () {
       app.get("/a", redirectsTo(302, "http://localhost:3600/b"));
       app.get("/b", function (req, res) {
         res.write(JSON.stringify(req.headers));
@@ -1183,69 +1269,274 @@ describe("follow-redirects", function () {
           assert.equal(body.host, "localhost:3600");
         });
     });
-  });
 
-  describe("when the client passes an Authorization header", function () {
-    it("keeps the header when redirected to the same host", function () {
-      app.get("/a", redirectsTo(302, "/b"));
-      app.get("/b", function (req, res) {
-        res.end(JSON.stringify(req.headers));
-      });
-
-      var opts = url.parse("http://localhost:3600/a");
-      opts.headers = {
-        authorization: "bearer my-token-1234",
-      };
-
-      return server.start(app)
-        .then(asPromise(function (resolve, reject) {
-          http.get(opts, resolve).on("error", reject);
-        }))
-        .then(asPromise(function (resolve, reject, res) {
-          res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
-        }))
-        .then(function (str) {
-          var body = JSON.parse(str);
-          assert.equal(body.host, "localhost:3600");
-          assert.equal(body.authorization, "bearer my-token-1234");
-        });
-    });
-
-    it("keeps the header when redirected to the same host via header", function () {
+    it("uses the location host if redirect host is the same", function () {
       app.get("/a", redirectsTo(302, "http://localhost:3600/b"));
       app.get("/b", function (req, res) {
-        res.end(JSON.stringify(req.headers));
+        res.write(JSON.stringify(req.headers));
+        req.pipe(res); // will invalidate JSON if non-empty
       });
 
-      var opts = url.parse("http://127.0.0.1:3600/a");
-      opts.headers = {
-        host: "localhost",
-        authorization: "bearer my-token-1234",
-      };
-
       return server.start(app)
         .then(asPromise(function (resolve, reject) {
+          var opts = url.parse("http://127.0.0.1:3600/a");
+          opts.headers = { hOsT: "localhost:3600" };
           http.get(opts, resolve).on("error", reject);
         }))
         .then(asPromise(function (resolve, reject, res) {
+          assert.deepEqual(res.statusCode, 200);
+          assert.deepEqual(res.responseUrl, "http://localhost:3600/b");
           res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
         }))
         .then(function (str) {
           var body = JSON.parse(str);
           assert.equal(body.host, "localhost:3600");
-          assert.equal(body.authorization, "bearer my-token-1234");
         });
     });
 
-    it("drops the header when redirected to a different host", function () {
-      app.get("/a", redirectsTo(302, "http://127.0.0.1:3600/b"));
+    it("uses the existing host header if redirect host is relative", function () {
+      app.get("/a", redirectsTo(302, "/b"));
+      app.get("/b", function (req, res) {
+        res.write(JSON.stringify(req.headers));
+        req.pipe(res); // will invalidate JSON if non-empty
+      });
+
+      return server.start(app)
+        .then(asPromise(function (resolve, reject) {
+          var opts = url.parse("http://127.0.0.1:3600/a");
+          opts.headers = { hOsT: "localhost:3600" };
+          http.get(opts, resolve).on("error", reject);
+        }))
+        .then(asPromise(function (resolve, reject, res) {
+          assert.deepEqual(res.statusCode, 200);
+          assert.deepEqual(res.responseUrl, "http://localhost:3600/b");
+          res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+        }))
+        .then(function (str) {
+          var body = JSON.parse(str);
+          assert.equal(body.host, "localhost:3600");
+        });
+    });
+  });
+
+  [
+    "Authorization",
+    "Cookie",
+  ].forEach(function (header) {
+    describe("when the client passes an header named " + header, function () {
+      it("ignores it when null", function () {
+        app.get("/a", redirectsTo(302, "http://localhost:3600/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://127.0.0.1:3600/a");
+        opts.headers = { host: "localhost" };
+        opts.headers[header] = null;
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "localhost:3600");
+            assert.equal(body[header.toLowerCase()], undefined);
+          });
+      });
+
+      it("keeps the header when redirected to the same host", function () {
+        app.get("/a", redirectsTo(302, "/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://localhost:3600/a");
+        opts.headers = {};
+        opts.headers[header] = "the header value";
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "localhost:3600");
+            assert.equal(body[header.toLowerCase()], "the header value");
+          });
+      });
+
+      it("keeps the header when redirected to the same host via header", function () {
+        app.get("/a", redirectsTo(302, "http://localhost:3600/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://127.0.0.1:3600/a");
+        opts.headers = { host: "localhost:3600" };
+        opts.headers[header] = "the header value";
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "localhost:3600");
+            assert.equal(body[header.toLowerCase()], "the header value");
+          });
+      });
+
+      it("keeps the header when redirected to the same host via header", function () {
+        app.get("/a", redirectsTo(302, "http://localhost:3600/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://127.0.0.1:3600/a");
+        opts.headers = { host: "localhost:3600" };
+        opts.headers[header] = "the header value";
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "localhost:3600");
+            assert.equal(body[header.toLowerCase()], "the header value");
+          });
+      });
+
+      it("keeps the header when redirected to a subdomain", function () {
+        app.get("/a", redirectsTo(302, "http://sub.localhost:3600/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://localhost:3600/a");
+        opts.headers = {};
+        opts.headers[header] = "the header value";
+
+        // Intercept the hostname, as no DNS entry is defined for it
+        opts.beforeRedirect = function (options) {
+          assert.equal(options.hostname, "sub.localhost");
+          options.hostname = "localhost";
+        };
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "localhost:3600");
+            assert.equal(body[header.toLowerCase()], "the header value");
+          });
+      });
+
+      it("drops the header when redirected to a different host (same hostname and different port)", function () {
+        app.get("/a", redirectsTo(302, "http://localhost:3600/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://127.0.0.1:3600/a");
+        opts.headers = { host: "localhost" };
+        opts.headers[header] = "the header value";
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "localhost:3600");
+            assert.equal(body[header.toLowerCase()], undefined);
+          });
+      });
+
+      it("drops the header when redirected to a different host", function () {
+        app.get("/a", redirectsTo(302, "http://127.0.0.1:3600/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://localhost:3600/a");
+        opts.headers = {};
+        opts.headers[header] = "the header value";
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "127.0.0.1:3600");
+            assert.equal(body[header.toLowerCase()], undefined);
+          });
+      });
+
+      it("drops the header when redirected from a different host via header", function () {
+        app.get("/a", redirectsTo(302, "http://127.0.0.1:3600/b"));
+        app.get("/b", function (req, res) {
+          res.end(JSON.stringify(req.headers));
+        });
+
+        var opts = url.parse("http://127.0.0.1:3600/a");
+        opts.headers = { host: "localhost" };
+        opts.headers[header] = "the header value";
+
+        return server.start(app)
+          .then(asPromise(function (resolve, reject) {
+            http.get(opts, resolve).on("error", reject);
+          }))
+          .then(asPromise(function (resolve, reject, res) {
+            res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
+          }))
+          .then(function (str) {
+            var body = JSON.parse(str);
+            assert.equal(body.host, "127.0.0.1:3600");
+            assert.equal(body[header.toLowerCase()], undefined);
+          });
+      });
+    });
+
+    it("keeps the header when redirected from HTTP to HTTPS", function () {
+      app.get("/a", redirectsTo(302, "https://localhost:3600/b"));
       app.get("/b", function (req, res) {
         res.end(JSON.stringify(req.headers));
       });
 
       var opts = url.parse("http://localhost:3600/a");
-      opts.headers = {
-        authorization: "bearer my-token-1234",
+      opts.headers = {};
+      opts.headers[header] = "the header value";
+
+      // Intercept the scheme
+      opts.beforeRedirect = function (options) {
+        assert.equal(options.protocol, "https:");
+        options.protocol = "http:";
       };
 
       return server.start(app)
@@ -1257,34 +1548,37 @@ describe("follow-redirects", function () {
         }))
         .then(function (str) {
           var body = JSON.parse(str);
-          assert.equal(body.host, "127.0.0.1:3600");
-          assert.equal(body.authorization, undefined);
+          assert.equal(body[header.toLowerCase()], "the header value");
         });
     });
 
-    it("drops the header when redirected from a different host via header", function () {
-      app.get("/a", redirectsTo(302, "http://127.0.0.1:3600/b"));
+    it("drops the header when redirected from HTTPS to HTTP", function () {
+      app.get("/a", redirectsTo(302, "http://localhost:3601/b"));
       app.get("/b", function (req, res) {
         res.end(JSON.stringify(req.headers));
       });
 
-      var opts = url.parse("http://127.0.0.1:3600/a");
-      opts.headers = {
-        host: "localhost",
-        authorization: "bearer my-token-1234",
+      var opts = url.parse("https://localhost:3601/a");
+      opts.ca = ca;
+      opts.headers = {};
+      opts.headers[header] = "the header value";
+
+      // Intercept the scheme
+      opts.beforeRedirect = function (options) {
+        assert.equal(options.protocol, "http:");
+        options.protocol = "https:";
       };
 
-      return server.start(app)
+      return server.start(httpsOptions(app))
         .then(asPromise(function (resolve, reject) {
-          http.get(opts, resolve).on("error", reject);
+          https.get(opts, resolve).on("error", reject);
         }))
         .then(asPromise(function (resolve, reject, res) {
           res.pipe(concat({ encoding: "string" }, resolve)).on("error", reject);
         }))
         .then(function (str) {
           var body = JSON.parse(str);
-          assert.equal(body.host, "127.0.0.1:3600");
-          assert.equal(body.authorization, undefined);
+          assert.equal(body[header.toLowerCase()], undefined);
         });
     });
   });
